# 
# XTension Plugin for API version 1.0
# this plugin controls the 6 channel dimmer and GPIO board kit from XTension
# http://MacHomeAutomation.com/
# c2017 by james@sentman.comm
#
#  this is also released as example code for creating your own plugins
#  you may re-use or re-purpose any part of this code for your own 
#  XTension plugin creation.
#
#  full XTension plugin documentation is available at:
#  http://machomeautomation.com/doku.php/plugins
#
#


#
# main plugin file must import these 2 modules
#
from XTension import *
from XTensionHelpers import *


#
# These are constants for the 4 unit tags that I define in my info.json file
#
tagPWM 			= 'PWM'
tagGPIO 		= 'GPIO'
tagRGB			= 'RGB'
tagRegister		= 'REG'
tagTemperature	= 'TEMP1'

#
#  C H E C K   D E P E N D E N C I E S
#
#  this is the first entry point into your script as it's being loaded.
#  you must return True from this event in order to have the loading continue.
#  this handler is required. You must return True to continue loading.
#   
def XTCheckDependencies():
	return True
	
#
#  D E B U G   M O D E   C H A N G E D
#
#  whenever debug mode is turned on or off this event will be called with the new value
#  you can access the global xtDebugMode at any time, but it is sometimes valuable to print
#  some more info when the mode it manually turned on or off, you can do that here.
#  this handler is optional.
#
def XTDebugModeChanged( newValue):
	writeLog( "the debug value is available available using the debugMode module variable")
	writeLog( "but you can also get an event when the mode has changed in case you wish to log")
	writeLog( "the current state at that time")
	
#
#  N E W   U N I T S
#
#  This event is called during startup, whenever a unit is added or it's configuration is saved or the
#  database is saved. myUnitList is a tuple of the uniqueID's of all the units assigned to
#  your interface. You can save off that list to a global or at anytime you can walk through
#  the 2 global dictionaries that are maintained for finding your units:
#  XTUnitsById[ theUniqueIdString] and
#  XTUnitsByAddress[ theUnitTypeTag + theUnitAddress]
#
#  this handler is optional
#
def XTNewUnits( myUnitList):

	writeLog( "There are " + str( len( myUnitList)) + " units assigned to this interface.")
#	print( myUnitList)
	

#
#	C O N F I G U R A T I O N   C H A N G E D
#
#	called when the device configuration has loaded the first time and again
#   when ever the interface settings are saved. Nothing has necessarily changed, the
#   interface window was just opened and saved.
#
#   the parameter newConfigInfo is a python dictionary that contains all the keyed values
#   for the dynamic user interface in the Interface configuration dialog as well as most
#   values from the built in settings like the serial port name or TCP address and port.
#   see the Constants for accessing information in the Interface level communications 
#   dictionary section of the XTensionHelpers.py include file for what you can get out of
#   this beyond your own dynamic interface keys.
#
#	if your interface needs to be restarted in order to accept these changes then you 
#   should call XTStop(). For most changes to IP addresses or ports the entire interface
#   needs to be restarted anyway.
#
#   this handler is optional
#

def XTConfigurationChanged( newConfigInfo):
#	create a global to save off the configuration so that we can refer to it
#	later as there is no call to get it back after this event.	
	global myConfig
	myConfig = newConfigInfo
#
# 	in my dynamic interface I have a popup menu with a tag of 'pwm1' I can retrieve the
# 	value of the selection by using this dictionary and make any changes to the device
#
	writeLog( "pwm1 selected tag: " + myConfig[ 'pwm1'])
#
# 	as a debugging aid you can print the entire dictionary here also to see what is
# 	in it exactly. You'll want to not do this in a released plugin
#
	print( myConfig)
	
#
#	P O R T   O P E N E D
#
#  	when the communications is ready you'll receive this event
#  	if you have the communications set to "none" or have set the 
#  	"scriptWillHandleComms" key in the info file set to true then this same event
# 	will fire as the ready event telling you that everything is ready and you can setup
#   and manage any communications that you have ready. If the connection is lost
#   the interface process will quit and restart allowing for a clean reset later
#   unless the communications are incoming. Incoming communications are handled differently
#   and as of this writing unimplemented.
#
#	any initial setup or talking to a device you can do here as the pipe is now ready for
#	the XTSendData command and you may begin receiving XTDataAvailable events at any time
#	after this call.
#
def XTPortOpened():

	writeLog( "the script has been alerted to the fact that there is a connection")	
	
#	
#	Now that I have a pipe to the dimmer I being my ping timer so that I can regularly talk
#	to it and verify that it's there and responding.
#
	writeLog( "starting ping timer")

#
#	create a global or a module level variable to hold the reference to the timer object
#	if the timer object goes out of scope you will stop getting your timer events
#
	global pingTimer 

#
#	Here I am creating a timer that will call back to a function in this script called
#	"pingTimerCallback" every 10 seconds until I set the global to something else.
#	I am also including a "tag" value of 0 though thats not used for anything in this
# 	simple case.
#
	pingTimer = XTMakeTimer( xtTimerModeMultiple, 10000, "pingTimerCallback", 0)
	
			


#
#	P I N G   T I M E R   C A L L B A C K
#
# 	I create a timer callback later to regularly ping the board once the port is opened.
# 	This is set to true when I send a ping to the board and set to false when I receive
# 	it's response. If I get another timer event to send a ping and find this was not
# 	reset to false then I log that the connection was dropped and attempt to reconnect.
#
pingOutstanding = False

def pingTimerCallback( theTag):
	
	# if we get called here after having not received a response to the last ping then we
	# should quit and restart
	global pingOutstanding
	if pingOutstanding == True:
		XTWriteLog( "no ping received from the device, attempt to reconnect", xtColorRed)
		pingOutstanding = False
		XTReconnect()
	else:
		pingOutstanding = True
		XTSendData( "PING\r")




#
#  S H U T D O W N
#	
#	called when the interface is being disabled in XTension or when XTension is being
#	quit and all it's interfaces need to shutdown. 
#	This allows you do to a clean disconnection with the device and send any last
#	commands that you might need to send. The interface host will wait for a few seconds
#	after calling this and poll the device socket or port to make sure any data you send 
#	will arrive before shutting down the pipe. There is a reasonable timeout for this and
#	if you don't return within 30 seconds your interface will be forced to quit.
#
def shutdown():
	writeLog( "the script has been alerted to the fact that the interface is being quit")
	
	# remove the reference to the pingTimer so that it will stop calling back
	# this isn't strictly necessary at shutdown since all the objects will be cleaned up
	# momentarily anyway, but is placed here as an example.
	global pingTimer
	pingTimer = {}






#
#	D A T A   A V A I L A B L E
#
#	command format from the dimmer module is very simple. 
#	if the line doesn't contain an "=" then the line is a log line and will be
# 	sent to the log. If it does contain an "=" then it's a command from the dimmer
# 	that needs to be properly forwarded to XTension. Here I create the appropriate
#	XTension command dictionary and figure out the appropriate unit device tag to use
#	so that the unit in XTension receives the appropriate new value or state.
#
#	in this case the device sends each line delineated by a carriage return so I placed
#	that as the packet boundary in the info.json file. So I will get a dataAvailable
#	event for each packet. If you don't have a protocol that can be parsed that way
# 	then managing a buffer of data and parsing it into individual packets is up to your
#	script here. Without a boundary setting this calls whenever any data is received. It 
#	may contain only a part of a longer packet, or it may contain several packets.
#

def XTDataAvailable( theData):

#	bring in the xtDebugMode global so that we can write the data received to the log
# 	if debug mode is on. This is very helpful for debugging and initial development of a 
#	protocol handler.
	global xtDebugMode
	if xtDebugMode:
		writeLog( "in dataAvailable (" + theData + ")")

#
#	sometimes the device just sends a single carriage return or I get an empty event
# 	we want to just ignore that.
#
	if theData == '':
		return
		
	if theData == '\r':
		return


#
#		anything that doesn't have an equal sign in it should just be logged
#
	if "=" not in theData:
		writeLog( theData)
		return
		
#
#		anything left is a command and needs to be forwarded to XTension
#		according to it's address we calculate the address suffix
#

	[theAddress, theValue] = theData.split( "=", 2)
	
	if theAddress == '':
		return

	
	#writeLog( "address=" + theAddress + " value=" + theValue)
	
#
#		D I M M E R   C H A N N E L S
#
#		if the address is one of the simple numbers [1..6] then its a dimmer channel value
#		but in order to get a range of [1, 2, 3, 4, 5, 6] we have to ask for a range( 1, 7)
#
#		a local change or an update to verify that a command was received for a channel value
#		change looks like this:
#		1=150\r
#		the channel number, an equals sign and the value
#
	if theAddress.isdigit() and int( theAddress) in range( 1, 7):
	
#
#	An XTension command is a python dictionary with keys that describe the unit and the 
#	data that we wish to send. 
# 	Create an empty dictionary and then add the necessary keys to it.
#
		xtCommand = {}
#
#	xtKeyAddress is absolutely required for any unit command
#	in this case the address is the left portion of the datapacket we received so
#	just pass that through as the address
#
		xtCommand[ xtKeyAddress] = theAddress
#
#	xtKeyTag is absolutely required for any unit command
# 	in this case we know we're talking to one of the dimmer channels based on the
#	left portion of the packet so we can set it to the tagPWM as defined in the unit
#	type definitions in our info.json file.
#
		xtCommand[ xtKeyTag] = tagPWM
#
#	xtKeyUpdateOnly is not required, but it allows us to send a command to a unit and not
#	have the unit actually do anything if the value is the same. In the case of this device
# 	when you control a dimmer channel it will begin fading to that value at the current ramp
#	rate and when it reaches that value it will send a new command down the pipe. If we set 
#	this flag then if the value is new it will receive that and if the value is a verification
#	of the last command then we don't have to do anything more complicated to keep track of
#	an outstanding command and what to do with a new reception than to set this flag.
#
		xtCommand[ xtKeyUpdateOnly] = xtTrue

#
#	xtKeyCommand is required
#
# 	for this device the command is the right hand side of the equal sign in the packet
#	we are parsing. For a dimmer channel the device can send ON OFF or a numerical value
# 	so we need to transfer those possibilities into the XTension command.
#
		if theValue == 'ON':
			xtCommand[ xtKeyCommand] = xtCommandOn
		elif theValue == 'OFF':
			xtCommand[ xtKeyCommand] = xtCommandOff
		elif theValue.isdigit():
#
#	If we are a value setting then we need to include the command but also
#	the xtKeyValue to tell XTension what value it's being set to.
#
	
			xtCommand[ xtKeyCommand] = xtCommandSetValue
			
#
#				values at the dimmer are 0-255 but XTension wants 0-100
#				so calculate this and round to the nearest integer
#
			
			xtCommand[ xtKeyValue] = str( int( round( float( theValue) / 2.55)))
		else:
			writeLog( "an unknown dimmer value was received: " + theAddress + "=" + theValue, xtColorRed)
			return
	
		XTSendCommand( xtCommand)




#
#		   G P I O  
#
#	 we've got a GPIO command so forward that through with the appropriate prefix/tag
#	gpio changes from the device start with GPIO so we can parse out the left portion of 
#	the packet based on that.
#
	elif theAddress.startswith( "GPIO"):
	
		xtCommand = {}
		xtCommand[ xtKeyAddress] = theAddress
#
#	Here also set the appropriate unit device tag from our info.json file
#
		xtCommand[ xtKeyTag] = tagGPIO
		xtCommand[ xtKeyUpdateOnly] = xtTrue
		
		if theValue == 'ON':
			xtCommand[ xtKeyCommand] = xtCommandOn
			XTSendCommand( xtCommand)
		elif theValue == 'OFF':
			xtCommand[ xtKeyCommand] = xtCommandOff
			XTSendCommand( xtCommand)
		elif theValue.isdigit():
			xtCommand[ xtKeyCommand] = xtCommandSetValue
			xtCommand[ xtKeyValue] = theValue
			XTSendCommand( xtCommand)
		else:
#
#			the gpio channels will also respond with their address and their
#			setting of input/output_on/output_off/counter/analog so need to handle that too
#
			
			writeLog( theAddress + " is set to " + theValue, xtColorGreen)


#
#   	T E M P E R A T U R E
#
#	one wire temperature addresses begin with a T
#

	elif theAddress.startswith( "T"):
	
		xtCommand = {}
		xtCommand[ xtKeyAddress] = theAddress
		xtCommand[ xtKeyTag] = tagTemperature
#
#	also set the update only flag so that we can send every value received without
#	having to compare against the last value set for this sensor and deciding if it
#	is different or not. With this flag set XTension will ignore the value unless it is 
#	different. However the last Message Received date will update with the reception
#	letting you use the idle timeout error in the advanced tab of the unit to show if a 
#	device has stopped responding.
#
		xtCommand[ xtKeyUpdateOnly] = xtTrue
		xtCommand[ xtKeyCommand] = xtCommandSetValue
		xtCommand[ xtKeyValue] = theValue
		
		XTSendCommand( xtCommand)
		
#
#		P I N G   R E P L Y 
#
# 	After I send a ping to the device it will reply with PING=ON which will end up here
#	All thats really necessary here is that I turn off ping outstanding
# 	however I am also sending a NoOp command to a register unit with the address of "PING"
#	nothing will be logged but the last message received date of that PING unit will be updated
#	and I will be able to tell if the pinging has stopped. This will also update the
#	last message received date of our interface and keep XTension from having to ping
#	us to make sure we're still alive.
#

	elif theAddress == "PING":
#
#	turn off the ping outstanding flag so that it won't send an error
#
		global pingOutstanding

		pingOutstanding = False
		xtCommand = {}
		xtCommand[ xtKeyAddress] = "PING"
		xtCommand[ xtKeyTag] = tagRegister
		xtCommand[ xtKeyCommand] = xtCommandNoOp

		XTSendCommand( xtCommand)
			
	else:
	
#
#	a command was received that didn't match any of our expected inputs. Might be an error
# 	or something truncated by the tiny buffer in the device so we just log it so we 
#	can have a look at it and discover what happened.
#
	
		writeLog( "unprocessed command: " + theData, xtColorRed)
	








#
# 	U N I T   C O M M A N D
#	
#	this is called by XTension when a user or a script has made a change to a unit that you 
#	There may be other commands sent to the script through this denoting other information
#	see the constants in the XTensionHelpers.py file for more info on what might be received
#	your script should be prepared to deal with commands that might not be handled or expected
#	
#	theCommand received below is basically the same as the commands you created to send
# 	to XTension in the XTDataAvailable event. It's a dictionary with keyed values so you
#	can decide how to act upon the unit command.
#
	
def XTUnitCommand( theCommand):
#
#	Every command will have the xtKeyCommand key set
#
	workCommand = theCommand[ xtKeyCommand]
#
#	top level commands like a ping request will not have a tag or an address set, so in this case
# 	I am using the .get( key, default) call to get the key and set a default value if
#	it's not set to anything. Trying to get the key without this will result in a python
#	key not found error. You could also not pull the tag and address until you have
#	verified that the command you have received is one that guarantees those values are
#	present.
#
	workTag = theCommand.get( xtKeyTag, "none")
#	workAddress = theCommand.get( xtKeyAddress, "none")

	
	
# 	this is a PWM dimmer object so the command might be ON/OFF or a level
# 	in this case XTension will always include the value for an ON

	if workTag == tagPWM:
#
#	this is definitely a command for a unit so we can get the address here
#
		workAddress = theCommand[ xtKeyAddress]
	
#
# 	set value or on or off commands may include a ramp rate
#	if the xtKeyRate key is in the command then we want to send the
#	command to the dimer that sets that value for the next fade process
#	if no ramp rate is included then we want to set the value to 0 in the device
#	which causes it to load it's current default ramp rate into the value.
#
		if xtKeyRate in theCommand:
			XTSendData( "TEMPRAMP=" + str( theCommand[ xtKeyRate]) + "\r")
		else:
			XTSendData( "TEMPRAMP=0\r")
			
#
#	an ON command was received. If the unit in XTension was configured as simulated or smart
#	then the ON command will also include a value which we should send instead of the ON
#	command. Remembering to convert the 0-100 percent value from XTension into a 0-255
#	value for the dimmer. A future version will let you set a min/max value for a unit
#	so that you could use the entire dimming resolution of such devices.
#
		if workCommand == xtCommandOn:
			
			if xtKeyValue in theCommand:
				XTSendData( workAddress + "=" + str( int(int(theCommand[ xtKeyValue]) * 2.55)) + "\r")
			else:
				XTSendData( workAddress + "=ON\r")
				
#
#	OFF command simple nothing else to set
#
		if workCommand == xtCommandOff:
			XTSendData( workAddress + '=OFF\r')

#
#	SET VALUE command will have the xtKeyValue set
#
		if workCommand == xtCommandSetValue:
			XTSendData( workAddress + "=" + str( int( int( theCommand[ xtKeyValue]) * 2.55)) + "\r")
			
#
# 	a command for a GPIO pin. These can only be on or off
#
	elif workTag == tagGPIO:

		if workCommand == xtCommandOn:
			XTSendData( theCommand[ xtKeyAddress] + "=ON\r")
		elif workCommand == xtCommandOff:
			XTSendData( theCommand[ xtKeyAddress] + "=OFF\r")

#
#	a command for the RGB unit! This may have color data included and we can therefore
#	send the color command to the device to show the appropriate color
#
	elif workTag == tagRGB:
		
#
#	this may also include a ramp rate, so handle that if it's there
#		
		if xtKeyRate in theCommand:
			XTSendData( "TEMPRAMP=" + str( theCommand[ xtKeyRate]) + "\r")
		else:
			XTSendData( "TEMPRAMP=0\r")
			
#
#	For an ON or a set Value the color will always be present for a color capable device
#
		if workCommand == xtCommandOn or workCommand == xtCommandSetValue:
			XTSendData( "COLOR=" + theCommand[ xtKeyColor] + "\r")
		elif workCommand == xtCommandOff:
			XTSendData( "COLOR=000000\r")
			
			

			
			



#
#  D Y N A M I C   I N T E R F A C E   H A N D L E R S
#

#
# if you first create a simple dynamic interface and set the action of a button to this
# it will print the entire dictionary that is sent so that you can see exactly what is 
# being send and what is available from the dynamic interfaces. This function is not
# used anywhere in my own interfaces anymore, it is just an example and you can use it
# to get a handle on the feedback from dynamic controls
#
def testButton( rawData, positionalData):
#
#  print the entire dictionary to the log just for seeing it in it's raw python form
#  the print command is better than the write log command in this case because it will 
#  convert the more complex object type of a dictionary into a more or less human readable
#  format. XTWriteLog requires that the items being logged are strings.
#
	print( rawData)

#
#  walk the keys from the dictionary and write them all to the log as well
#

	keys = rawData.keys()
	for workKey in keys:
		print( "(" + workKey + ") = (" + rawData[ workKey] + ")")


#
# This is an interface "action" handler and also a menu handler. There are several
# buttons in the dynamic UI from the info.json file that define this action handler
# if you select to stop color loop in the menu or click any of the interface buttons
# this handler will be called in the script. 
# 
# use the XTSendData command to send the command to the dimmer to stop the color loop
#
def stopColorLoop( theDict, theList):
	XTWriteLog( "Stop color loop")
	XTSendData( "LOOP=0\r")


#
# This is an interface "action" handler. When the Start Color Loop menu action or 
# button is clicked this handler is run. The data from the dynamic controls is passed
# to me in "theDict" and I can pull out the value of any of my controls in the UI at the
# time the user clicked. In this example I am getting the 'loopspeed' key from the dynamic
# slider in the UI and sending that to the device first. Then I am getting the value of the
# unit and using that to start the loop with the current brightness of the unit.
#
# this demonstrates how much info is available from the UI at the time these are called
#
def startColorLoop( theDict, theList):
	XTWriteLog( "Start color loop")
	workLoopSpeed = theDict[ 'loopSpeed']
	XTSendData( "LOOPSPEED=" + workLoopSpeed + "\r")

	# the current value of the unit who's interface was used to click this value is always
	# added to this dict as well even though it is not technically part of the UI.
	myLevel = theDict[ xtKeyValue]
	
	# the level in XTension is from 0 to 100 so here we need to make it from 0 to 255
	myLevel = str( int( float( myLevel) * 2.55))
	XTSendData( "LOOP=" + myLevel + "\r")
	
	
#
# this method is the action handler from the loop speed slider. Whenever the user
# changes the value of the slider and the interface is running this will be called
# and I will resend the loopspeed command to the device with the new speed.
#
def colorLoopSpeedSlider( theDict, theList):
	theSpeed = theDict[ 'loopSpeed']
	XTSendData( "LOOPSPEED=" + theSpeed + "\r")
	
	

	
	
	




















	